import serial

from .exceptions import UsbIssError
from . import defs

# In Py2, bytes means str, and there's no immutable byte array defined.
# Use bytearray instead - this is mutable, but otherwise equivalent to
# Python3's bytes.
if isinstance(bytes(), str):
    bytes = bytearray

SERIAL_OPTS = {
    'baudrate': 9600,
    'parity': serial.PARITY_NONE,
    'bytesize': serial.EIGHTBITS,
    'stopbits': serial.STOPBITS_ONE,
    'xonxoff': False,
    'timeout': 0.5,
}


class Driver(object):
    """
    Internal serial port driver. Don't use this class directly.
    """
    def __init__(self):
        self._serial = None

    def open(self, port):
        self._serial = serial.Serial(port=port, **SERIAL_OPTS)
        return self

    def close(self):
        if self._serial is not None:
            self._serial.close()
            self._serial = None

    def write_cmd(self, command, data=None):
        if self._serial is None:
            raise UsbIssError("Serial port has not been opened")

        if data is None:
            data = []
        self._serial.write(bytes([command] + data))

    def read(self, byte_count):
        if self._serial is None:
            raise UsbIssError("Serial port has not been opened")

        data = list(self._serial.read(byte_count))
        if len(data) != byte_count:
            raise UsbIssError(
                "Expected %d bytes, but %d received" % (byte_count, len(data)))
        return data

    def check_i2c_ack(self):
        """For I2C, any non-zero code means ACK"""
        data = self.read(1)
        if data[0] == defs.ResponseCode.NACK.value:
            raise UsbIssError("Received NACK instead of ACK")

    def check_ack(self):
        data = self.read(1)
        if data[0] != defs.ResponseCode.ACK.value:
            raise UsbIssError("Received 0x%02X instead of ACK" % data[0])

    def check_ack_error_code(self, error_enum):
        data = self.read(2)
        if data[0] != defs.ResponseCode.ACK.value:
            raise UsbIssError(
                "Received %s [0x%02X, 0x%02X] instead of ACK"
                % (error_enum(data[1]), data[0], data[1]))
        return data[1]


class DummyDriver(object):
    """
    Dummy Driver object, used for testing.
    """
    def open(self, _):
        return self

    def close(self):
        pass

    def write_cmd(self, command, data=None):
        pass

    def read(self, byte_count):
        return list(range(byte_count))

    def check_i2c_ack(self):
        pass

    def check_ack(self):
        pass

    def check_ack_error_code(self, error_enum):
        return 0



from .exceptions import UsbIssError
from . import defs

I2C_RD = 0x01


class I2C(object):
    """
    Use the USB_ISS module to perform I2C accesses.

    Example:
        ::

            from usb_iss import UsbIss, defs

            # Configure I2C mode

            iss = UsbIss()
            iss.open("COM3")
            iss.setup_i2c()

            # Write and read back some data

            iss.i2c.write(0xC4, 0, [0, 1, 2]);
            data = iss.i2c.read(0xC4, 0, 3)

            print(data)
            # [0, 1, 2]
    """
    def __init__(self, drv):
        self._drv = drv

    def write(self, address, register, data):
        """
        Write multiple bytes to a device with a one-byte internal register
        address. The majority of devices will be written to using this method.
        This is an alias for the write_ad1 method.

        Args:
            address (int): I2C address of the device.
            register (int): Internal register address to write
                (0x00 - 0xFF).
            data (list of int): List of bytes to write to the device.
        """
        self.write_ad1(address, register, data)

    def read(self, address, register, byte_count):
        """
        Read multiple bytes from a device with a one-byte internal register
        address. The majority of devices will be read from using this method.
        This is an alias for the read_ad1 method.

        Args:
            address (int): I2C address of the device.
            register (int): Internal register address to read (0x00 - 0xFF).
            byte_count (int): Number of bytes to read.
        Returns:
            list of int: List of bytes read from the device.
        """
        return self.read_ad1(address, register, byte_count)

    def write_single(self, address, data_byte):
        """
        Write a single byte to an I2C device.

        Args:
            address (int): I2C address of the device.
            data_byte (int): Data byte to write to the device.
        """
        address = address & ~I2C_RD
        self._drv.write_cmd(defs.I2CCommand.I2C_SGL.value,
                            [address, data_byte])
        self._drv.check_i2c_ack()

    def read_single(self, address):
        """
        Read a single byte from an I2C device.

        Args:
            address (int): I2C address of the device.
        Returns:
            int: Data byte read from the device.
        """
        address = address | I2C_RD
        self._drv.write_cmd(defs.I2CCommand.I2C_SGL.value, [address])
        return self._drv.read(1)[0]

    def write_ad0(self, address, data):
        """
        Write multiple bytes to a device without internal register addressing,
        or where the internal register address does not require resetting.

        Args:
            address (int): I2C address of the device.
            data (list of int): List of bytes to write to the device.
        """
        address = address & ~I2C_RD
        self._drv.write_cmd(defs.I2CCommand.I2C_AD0.value,
                            [address, len(data)] + data)
        self._drv.check_i2c_ack()

    def read_ad0(self, address, byte_count):
        """
        Read multiple bytes from a device without internal register addressing,
        or where the internal register address does not require resetting.

        Args:
            address (int): I2C address of the devie.
            byte_count (int): Number of bytes to read.
        Returns:
            list of int: List of bytes read from the device.
        """
        address = address | I2C_RD
        self._drv.write_cmd(defs.I2CCommand.I2C_AD0.value,
                            [address, byte_count])
        return self._drv.read(byte_count)

    def write_ad1(self, address, register, data):
        """
        Write multiple bytes to a device with a one-byte internal register
        address.

        Args:
            address (int): I2C address of the device.
            register (int): Internal register address to write (0x00 - 0xFF).
            data (list of int): List of bytes to write to the device.
        """
        if len(data) > defs.I2C_AD1_MAX_WRITE_BYTE_COUNT:
            raise UsbIssError(
                "Attempted to write %d bytes, maximum is %d" %
                (len(data), defs.I2C_AD1_MAX_WRITE_BYTE_COUNT))
        address = address & ~I2C_RD
        self._drv.write_cmd(defs.I2CCommand.I2C_AD1.value,
                            [address, register, len(data)] + data)
        self._drv.check_i2c_ack()

    def read_ad1(self, address, register, byte_count):
        """
        Read multiple bytes from a device with a one-byte internal register
        address.

        Args:
            address (int): I2C address of the device.
            register (int): Internal register address to read (0x00 - 0xFF).
            byte_count (int): Number of bytes to read.
        Returns:
            list of int: List of bytes read from the device.
        """
        if byte_count > defs.I2C_AD1_MAX_READ_BYTE_COUNT:
            raise UsbIssError(
                "Attempted to read %d bytes, maximum is %d" %
                (byte_count, defs.I2C_AD1_MAX_READ_BYTE_COUNT))
        address = address | I2C_RD
        self._drv.write_cmd(defs.I2CCommand.I2C_AD1.value,
                            [address, register, byte_count])
        return self._drv.read(byte_count)

    def write_ad2(self, address, register, data):
        """
        Write multiple bytes to a device with a two-byte internal register
        address.

        Args:
            address (int): I2C address of the device.
            register (int): Internal register address to write
                (0x0000 - 0xFFFF).
            data (list of int): List of bytes to write to the device.
        """
        if len(data) > defs.I2C_AD2_MAX_WRITE_BYTE_COUNT:
            raise UsbIssError(
                "Attempted to write %d bytes, maximum is %d" %
                (len(data), defs.I2C_AD2_MAX_WRITE_BYTE_COUNT))
        address = address & ~I2C_RD
        reg_high = register >> 8
        reg_low = register & 0xFF
        self._drv.write_cmd(defs.I2CCommand.I2C_AD2.value,
                            [address, reg_high, reg_low, len(data)] + data)
        self._drv.check_i2c_ack()

    def read_ad2(self, address, register, byte_count):
        """
        Read multiple bytes from a device with a two-byte internal register
        address.

        Args:
            address (int): I2C address of the device.
            register (int): Internal register address to read
                (0x0000 - 0xFFFF).
            byte_count (int): Number of bytes to read.
        Returns:
            list of int: List of bytes read from the device.
        """
        if byte_count > defs.I2C_AD2_MAX_READ_BYTE_COUNT:
            raise UsbIssError(
                "Attempted to read %d bytes, maximum is %d" %
                (byte_count, defs.I2C_AD2_MAX_READ_BYTE_COUNT))
        address = address | I2C_RD
        reg_high = register >> 8
        reg_low = register & 0xFF
        self._drv.write_cmd(defs.I2CCommand.I2C_AD2.value,
                            [address, reg_high, reg_low, byte_count])
        return self._drv.read(byte_count)

    def direct(self, data):
        """
        Send a custom I2C sequence to the device.
        See https://www.robot-electronics.co.uk/htm/usb_iss_i2c_tech.htm for a
        full set of examples.

        Args:
            data (list of defs.I2CDirect): List of
                :class:`~usb_iss.defs.I2CDirect` commands and data.

        Returns:
            list of int: List of bytes read from the device.

        Example:
            ::

                # Equivalent to iss.i2c.write_single(0x40, 0x55)
                iss.i2c.direct([
                    defs.I2CDirect.START,
                    defs.I2CDirect.WRITE2,
                    0x40,
                    0x55,
                    defs.I2CDirect.STOP,
                ]);

        """
        # Convert any I2CDirect items to the corresponding value
        def convert_to_value(byte):
            if isinstance(byte, defs.I2CDirect):
                return byte.value
            else:
                return byte
        bytes = [convert_to_value(byte) for byte in data]

        self._drv.write_cmd(defs.I2CCommand.I2C_DIRECT.value, bytes)
        bytes_to_read = self._drv.check_ack_error_code(defs.I2CDirectError)
        if bytes_to_read == 0:
            return []
        return self._drv.read(bytes_to_read)

    def test(self, address):
        """
        Check whether a device responds at the specified I2C addresss.

        Args:
            address (int): I2C address of the device.

        Returns:
            bool: True if the device responds with an ACK.
        """
        self._drv.write_cmd(defs.I2CCommand.I2C_TEST.value, [address])
        return self._drv.read(1) != [defs.I2CTestResponse.NO_DEVICE.value]



from . import defs
from .exceptions import UsbIssError
from .driver import Driver, DummyDriver
from .i2c import I2C
from .io import IO

# In Py2, bytes means str, and there's no immutable byte array defined.
# Use bytearray instead - this is mutable, but otherwise equivalent to
# Python3's bytes.
if isinstance(bytes(), str):
    bytes = bytearray


class UsbIss(object):
    """
    Main USB_ISS object.

    Example:
        ::

            from usb_iss import UsbIss, defs

            # Configure I2C mode

            iss = UsbIss()
            iss.open("COM3")
            iss.setup_i2c()

            # Write and read back some data

            iss.i2c.write(0xC4, 0, [0, 1, 2]);
            data = iss.i2c.read(0xC4, 0, 3)

            print(data)
            # [0, 1, 2]

    Attributes:
        i2c (:class:`i2c.I2C`): Attribute to use for I2C access.
        io (:class:`io.IO`): Attribute to use for pin IO access.

    """
    def __init__(self, dummy=False):
        self._drv = DummyDriver() if dummy else Driver()

        self.i2c = I2C(self._drv)
        self.io = IO(self._drv)

    def open(self, port):
        """
        Open the specified serial port for communication with the USB_ISS
        module.

        Args:
            port (str): Serial port to use for usb_iss communication.
        """
        self._drv.open(port)
        return self

    def close(self):
        """
        Close the serial port.
        """
        self._drv.close()

    def setup_i2c(self, clock_khz=400, use_i2c_hardware=True,
                  io1_type=defs.IOType.DIGITAL_INPUT,
                  io2_type=defs.IOType.DIGITAL_INPUT):
        """
        Issue a ISS_MODE command to set the operating mode to I2C.

        Args:
            clock_khz (int): I2C clock rate in kHz.
                See https://www.robot-electronics.co.uk/htm/usb_iss_tech.htm
                for a list of valid values.
            use_i2c_hardware (bool): Use the USB_ISS module's hardware I2C
                controller.
            io1_type (defs.IOType): IO1 mode
                (default: DIGITAL_INPUT).
            io2_type (defs.IOType): IO2 mode
                (default: DIGITAL_INPUT).
        """
        if clock_khz == 20:
            assert not use_i2c_hardware
            i2c_mode = defs.Mode.I2C_S_20KHZ.value
        elif clock_khz == 50:
            assert not use_i2c_hardware
            i2c_mode = defs.Mode.I2C_S_50KHZ.value
        elif clock_khz == 100:
            i2c_mode = (defs.Mode.I2C_H_100KHZ.value if use_i2c_hardware else
                        defs.Mode.I2C_S_100KHZ.value)
        elif clock_khz == 400:
            i2c_mode = (defs.Mode.I2C_H_400KHZ.value if use_i2c_hardware else
                        defs.Mode.I2C_S_400KHZ.value)
        elif clock_khz == 1000:
            assert use_i2c_hardware
            i2c_mode = defs.Mode.I2C_H_1000KHZ.value
        else:
            raise UsbIssError("Invalid clk_khz value")

        io_type = io1_type.value | (io2_type.value << 2)
        data = [defs.SubCommand.ISS_MODE.value, i2c_mode, io_type]
        self._drv.write_cmd(defs.CMD_USB_ISS, data)
        self._drv.check_ack_error_code(defs.ModeError)

    def setup_i2c_serial(self):
        raise NotImplementedError

    def setup_spi(self):
        raise NotImplementedError

    def setup_io(self):
        raise NotImplementedError

    def change_io(self):
        raise NotImplementedError

    def setup_serial(self):
        raise NotImplementedError

    def read_module_id(self):
        """
        Returns:
            int: The USB_ISS module ID (always 7).
        """
        self._drv.write_cmd(defs.CMD_USB_ISS,
                            [defs.SubCommand.ISS_VERSION.value])
        return self._drv.read(3)[0]

    def read_fw_version(self):
        """
        Returns:
            int: The USB_ISS firmware version.
        """
        self._drv.write_cmd(defs.CMD_USB_ISS,
                            [defs.SubCommand.ISS_VERSION.value])
        return self._drv.read(3)[1]

    def read_iss_mode(self):
        """
        Returns:
            defs.Mode: The current ISS_MODE operating mode.
        """
        self._drv.write_cmd(defs.CMD_USB_ISS,
                            [defs.SubCommand.ISS_VERSION.value])
        return defs.Mode(self._drv.read(3)[2])

    def read_serial_number(self):
        """
        Returns:
            str: The serial number of the attached USB_ISS module.
        """
        self._drv.write_cmd(defs.CMD_USB_ISS,
                            [defs.SubCommand.GET_SER_NUM.value])
        return bytes(self._drv.read(8)).decode('ascii')



from . import defs


class IO(object):
    """
    Use the USB_ISS module to perform IO accesses.
    """
    def __init__(self, drv):
        self._drv = drv

    def set_pins(self, io0, io1, io2, io3):
        """
        Set the digital output pins high or low. This command only operates on
        pins that have been configured as digital outputs. Digital or analogue
        inputs or pins that are used for I2C or serial are not affected.

        For each argument, set to 0 to drive low and 1 to drive high.

        Args:
            io0 (int): IO0 output value.
            io1 (int): IO1 output value.
            io2 (int): IO2 output value.
            io3 (int): IO3 output value.
        """
        data = (((io0 & 0x01) << 0) +
                ((io1 & 0x01) << 1) +
                ((io2 & 0x01) << 2) +
                ((io3 & 0x01) << 3))
        self._drv.write_cmd(defs.IOCommand.SET_PINS.value, [data])
        self._drv.check_ack()

    def get_pins(self):
        """
        Get the current state of all digital IO pins.

        Returns:
            list of int: List containing the current state of the four digital
            IO pins (0 = low, 1 = high).
        """
        self._drv.write_cmd(defs.IOCommand.GET_PINS.value)
        data = self._drv.read(1)[0]
        return [(data >> 0) & 0x01,
                (data >> 1) & 0x01,
                (data >> 2) & 0x01,
                (data >> 3) & 0x01]

    def get_ad(self, pin):
        """
        Get a ADC sample from the specified analogue input pin.

        This uses a 10-bit ADC, so the result is between 0-1023 for the voltage
        swing between VSS and VCC.

        Args:
            pin (int): Input pin to sample.

        Returns:
            int: Sample value returned by the ADC (0-1023).
        """
        self._drv.write_cmd(defs.IOCommand.GET_AD.value, [pin])
        data = self._drv.read(2)
        return (data[0] << 8) + data[1]



"""
Definitions defined at http://www.robot-electronics.co.uk/htm/usb_iss_tech.htm
"""
from enum import Enum

# USB-ISS command
CMD_USB_ISS = 0x5A


class ResponseCode(Enum):
    """
    Note: For I2C, NACK = 0x00, ACK = Non-zero.
    """
    NACK = 0x00
    ACK = 0xFF


class SubCommand(Enum):
    ISS_VERSION = 0x01
    ISS_MODE = 0x02
    GET_SER_NUM = 0x03


class Mode(Enum):
    IO_MODE = 0x00
    IO_CHANGE = 0x10
    I2C_S_20KHZ = 0x20
    I2C_S_50KHZ = 0x30
    I2C_S_100KHZ = 0x40
    I2C_S_400KHZ = 0x50
    I2C_H_100KHZ = 0x60
    I2C_H_400KHZ = 0x70
    I2C_H_1000KHZ = 0x80
    SPI_MODE = 0x90
    SERIAL = 0x01


class ModeError(Enum):
    UNKNOWN_COMMAND = 0x05
    INTERNAL_ERROR_1 = 0x06
    INTERNAL_ERROR_2 = 0x07


class IOType(Enum):
    OUTPUT_LOW = 0x00
    OUTPUT_HIGH = 0x01
    DIGITAL_INPUT = 0x02
    ANALOGUE_INPUT = 0x03


class SPIMode(Enum):
    TX_ACTIVE_TO_IDLE_IDLE_LOW = 0x90
    TX_ACTIVE_TO_IDLE_IDLE_HIGH = 0x91
    TX_IDLE_TO_ACTIVE_IDLE_LOW = 0x92
    TX_IDLE_TO_ACTIVE_IDLE_HIGH = 0x93


class I2CCommand(Enum):
    I2C_SGL = 0x53
    I2C_AD0 = 0x54
    I2C_AD1 = 0x55
    I2C_AD2 = 0x56
    I2C_DIRECT = 0x57
    I2C_TEST = 0x58


# Maximum number of bytes that can be read/written in a single command
I2C_AD1_MAX_WRITE_BYTE_COUNT = 60
I2C_AD1_MAX_READ_BYTE_COUNT = 60
I2C_AD2_MAX_WRITE_BYTE_COUNT = 59
I2C_AD2_MAX_READ_BYTE_COUNT = 64


class I2CDirect(Enum):
    START = 0x01
    RESTART = 0x02
    STOP = 0x03
    NACK = 0x04
    READ1 = 0x20
    READ2 = 0x21
    READ3 = 0x22
    READ4 = 0x23
    READ5 = 0x24
    READ6 = 0x25
    READ7 = 0x26
    READ8 = 0x27
    READ9 = 0x28
    READ10 = 0x29
    READ11 = 0x2A
    READ12 = 0x2B
    READ13 = 0x2C
    READ14 = 0x2D
    READ15 = 0x2E
    READ16 = 0x2F
    WRITE1 = 0x30
    WRITE2 = 0x31
    WRITE3 = 0x32
    WRITE4 = 0x33
    WRITE5 = 0x34
    WRITE6 = 0x35
    WRITE7 = 0x36
    WRITE8 = 0x37
    WRITE9 = 0x38
    WRITE10 = 0x39
    WRITE11 = 0x3A
    WRITE12 = 0x3B
    WRITE13 = 0x3C
    WRITE14 = 0x3D
    WRITE15 = 0x3E
    WRITE16 = 0x3F


class I2CDirectError(Enum):
    DEVICE_ERROR = 0x01
    BUFFER_OVERFLOW = 0x02
    BUFFER_UNDERFLOW = 0x03
    UNKNOWN_COMMAND = 0x04


class I2CTestResponse(Enum):
    """ Nonzero means device was found """
    NO_DEVICE = 0x00


class IOCommand(Enum):
    SET_PINS = 0x63
    GET_PINS = 0x64
    GET_AD = 0x65
